/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package crt

import (
	"context"
	"testing"

	api_v1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/client-go/informers"
	"k8s.io/client-go/kubernetes/fake"
	core_v1_listers "k8s.io/client-go/listers/core/v1"
	"k8s.io/client-go/tools/cache"
)

func setupSecretLister(client *fake.Clientset) core_v1_listers.SecretLister {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	infFactory := informers.NewSharedInformerFactory(client, 0)
	informer := infFactory.Core().V1().Secrets().Informer()
	lister := infFactory.Core().V1().Secrets().Lister()

	infFactory.Start(ctx.Done())
	cache.WaitForCacheSync(ctx.Done(), informer.HasSynced)
	return lister
}

func TestGetter(t *testing.T) {
	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns",
				Name:            "secret",
				UID:             "4cfffc85-1e86-4ebf-92f2-1d3fd61242f1",
				ResourceVersion: "4711",
			},
			Data: map[string][]byte{
				crtKey: []byte("crtSecret"),
				keyKey: []byte("keySecret"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "missing",
				Name:            "crtField",
				UID:             "3acf5e0a-1642-48f4-886f-4de83fd17f41",
				ResourceVersion: "0815",
			},
			Data: map[string][]byte{
				"foo":  []byte("fooVal"),
				keyKey: []byte("keySecret"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "missing",
				Name:            "keyField",
				UID:             "ea49b70e-c4ec-4b77-82e8-9c02e542f7af",
				ResourceVersion: "123456",
			},
			Data: map[string][]byte{
				crtKey: []byte("crtSecret"),
				"foo":  []byte("fooVal"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "empty",
				Name:            "crtField",
				UID:             "50dabc9a-d4e4-4db7-9b2a-563becd5fb06",
				ResourceVersion: "347360",
			},
			Data: map[string][]byte{
				crtKey: []byte{},
				keyKey: []byte("keySecret"),
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "empty",
				Name:            "keyField",
				UID:             "35df2a46-4299-4fe7-9d81-04314a0ef4b4",
				ResourceVersion: "956928",
			},
			Data: map[string][]byte{
				crtKey: []byte("crtSecret"),
				keyKey: []byte{},
			},
			Type: api_v1.SecretType("kubernetes.io/tls"),
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "invalid",
				Name:            "typeField",
				UID:             "688d59a2-6853-41bc-880c-62ea1e29e24a",
				ResourceVersion: "654321",
			},
			Data: map[string][]byte{
				crtKey: []byte("crtSecret"),
				keyKey: []byte("keySecret"),
			},
			Type: api_v1.SecretType("foo/bar"),
		},
	)
	lister := setupSecretLister(client)
	getter := NewGetter(lister)

	crt, key, suid, sversion, found, valid, err := getter.Get(
		"ns", "secret", "4cfffc85-1e86-4ebf-92f2-1d3fd61242f1", "4711")
	if err != nil {
		t.Errorf("err: got %+v want nil", err)
	}
	if !found {
		t.Error("found: got false want true")
	}
	if !valid {
		t.Error("found: got false want true")
	}
	if string(crt) != "crtSecret" {
		t.Errorf("crt: got %s want crtSecret", string(crt))
	}
	if string(key) != "keySecret" {
		t.Errorf("key: got %s want keySecret", string(key))
	}
	if suid != "4cfffc85-1e86-4ebf-92f2-1d3fd61242f1" {
		t.Errorf("suid: got %s want %s", suid,
			"4cfffc85-1e86-4ebf-92f2-1d3fd61242f1")
	}
	if sversion != "4711" {
		t.Errorf("sversion: got %s want 4711", sversion)
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"ns", "secret", "", "4711")
	if err != nil {
		t.Errorf("err: got %+v want nil", err)
	}
	if !found {
		t.Error("found: got false want true")
	}
	if !valid {
		t.Error("found: got false want true")
	}
	if string(crt) != "crtSecret" {
		t.Errorf("crt: got %s want crtSecret", string(crt))
	}
	if string(key) != "keySecret" {
		t.Errorf("key: got %s want keySecret", string(key))
	}
	if suid != "4cfffc85-1e86-4ebf-92f2-1d3fd61242f1" {
		t.Errorf("suid: got %s want %s", suid,
			"4cfffc85-1e86-4ebf-92f2-1d3fd61242f1")
	}
	if sversion != "4711" {
		t.Errorf("sversion: got %s want 4711", sversion)
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"ns", "secret", "4cfffc85-1e86-4ebf-92f2-1d3fd61242f1", "")
	if err != nil {
		t.Errorf("err: got %+v want nil", err)
	}
	if !found {
		t.Error("found: got false want true")
	}
	if !valid {
		t.Error("found: got false want true")
	}
	if string(crt) != "crtSecret" {
		t.Errorf("crt: got %s want crtSecret", string(crt))
	}
	if string(key) != "keySecret" {
		t.Errorf("key: got %s want keySecret", string(key))
	}
	if suid != "4cfffc85-1e86-4ebf-92f2-1d3fd61242f1" {
		t.Errorf("suid: got %s want %s", suid,
			"4cfffc85-1e86-4ebf-92f2-1d3fd61242f1")
	}
	if sversion != "4711" {
		t.Errorf("sversion: got %s want 4711", sversion)
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"ns", "secret", "", "")
	if err != nil {
		t.Errorf("err: got %+v want nil", err)
	}
	if !found {
		t.Error("found: got false want true")
	}
	if !valid {
		t.Error("found: got false want true")
	}
	if string(crt) != "crtSecret" {
		t.Errorf("crt: got %s want crtSecret", string(crt))
	}
	if string(key) != "keySecret" {
		t.Errorf("key: got %s want keySecret", string(key))
	}
	if suid != "4cfffc85-1e86-4ebf-92f2-1d3fd61242f1" {
		t.Errorf("suid: got %s want %s", suid,
			"4cfffc85-1e86-4ebf-92f2-1d3fd61242f1")
	}
	if sversion != "4711" {
		t.Errorf("sversion: got %s want 4711", sversion)
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"ns", "foo", "", "")
	if err == nil {
		t.Error("err: got nil want non-nil")
	}
	if found {
		t.Error("found: got true want false")
	}
	if !valid {
		t.Error("found: got false want true")
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"ns", "foo", "", "")
	if err == nil {
		t.Error("err: got nil want non-nil")
	}
	if found {
		t.Error("found: got true want false")
	}
	if !valid {
		t.Error("found: got false want true")
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"foo", "secret", "", "")
	if err == nil {
		t.Error("err: got nil want non-nil")
	}
	if found {
		t.Error("found: got true want false")
	}
	if !valid {
		t.Error("found: got false want true")
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"ns", "secret", "0815", "4711")
	if err != nil {
		t.Errorf("err: got %+v want nil", err)
	}
	if found {
		t.Error("found: got true want false")
	}
	if !valid {
		t.Error("found: got false want true")
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"ns", "secret", "4cfffc85-1e86-4ebf-92f2-1d3fd61242f1", "0815")
	if err != nil {
		t.Errorf("err: got %+v want nil", err)
	}
	if found {
		t.Error("found: got true want false")
	}
	if !valid {
		t.Error("found: got false want true")
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"missing", "crtField", "", "")
	if valid {
		t.Error("valid: got true want false")
	}
	if err == nil {
		t.Fatal("err: got nil want non-nil")
	}
	if err.Error() != "required field missing" {
		t.Errorf("err msg: want \"required field missing\" got %s",
			err.Error())
	}
	if !found {
		t.Error("found: got false want true")
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"missing", "keyField", "", "")
	if valid {
		t.Error("valid: got true want false")
	}
	if err == nil {
		t.Fatal("err: got nil want non-nil")
	}
	if err.Error() != "required field missing" {
		t.Errorf("err msg: want \"required field missing\" got %s",
			err.Error())
	}
	if !found {
		t.Error("found: got false want true")
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"empty", "crtField", "", "")
	if valid {
		t.Error("valid: got true want false")
	}
	if err == nil {
		t.Fatal("err: got nil want non-nil")
	}
	if err.Error() != "required field empty" {
		t.Errorf("err msg: want \"required field empty\" got %s",
			err.Error())
	}
	if !found {
		t.Error("found: got false want true")
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"empty", "keyField", "", "")
	if valid {
		t.Error("valid: got true want false")
	}
	if err == nil {
		t.Fatal("err: got nil want non-nil")
	}
	if err.Error() != "required field empty" {
		t.Errorf("err msg: want \"required field empty\" got %s",
			err.Error())
	}
	if !found {
		t.Error("found: got false want true")
	}

	crt, key, suid, sversion, found, valid, err = getter.Get(
		"invalid", "typeField", "", "")
	if valid {
		t.Error("valid: got true want false")
	}
	if err == nil {
		t.Fatal("err: got nil want non-nil")
	}
	if err.Error() != "invalid type value" {
		t.Errorf("err msg: want \"invalid type value\" got %s",
			err.Error())
	}
	if !found {
		t.Error("found: got false want true")
	}
}
