/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package k8s

import (
	"fmt"
	"os"
	"sync"

	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/pem"

	api_v1 "k8s.io/api/core/v1"
	utilruntime "k8s.io/apimachinery/pkg/util/runtime"
	"k8s.io/client-go/tools/record"
	"k8s.io/client-go/util/workqueue"

	"github.com/sirupsen/logrus"
)

// NamespaceWorker serves fanout of work items to workers for each
// namespace for which the controller is notified about a resource
// update. The NamespaceQueues object creates a new instance when it
// reads an item from a new namespace from its main queue. Thus each
// namespace is synced separately and sequentially, and all of the
// namespaces are synced in parallel.
type NamespaceWorker struct {
	namespace string
	log       *logrus.Logger
	queue     workqueue.RateLimitingInterface
	files     *pem.Files
	recorder  record.EventRecorder
	wg        *sync.WaitGroup
}

func (worker *NamespaceWorker) update(secret *api_v1.Secret) Status {
	if !worker.files.Have(secret.Namespace, secret.Name, string(secret.UID),
		secret.ResourceVersion) {
		return MakeNoop("pem file currently not stored")
	}

	if worker.files.Check(secret.Namespace, secret.Name, string(secret.UID),
		secret.ResourceVersion) {
		return MakeNoop("pem file already up to date")
	}

	if found, valid, err := worker.files.Write(
		secret.Namespace, secret.Name, string(secret.UID),
		secret.ResourceVersion); !found {
		return MakeRecoverable("pem file not found")
	} else if !valid {
		return MakeFatal("%v", err)
	} else if err != nil {
		if os.IsPermission(err) {
			return MakeFatal("%s", err)
		}
		return MakeRecoverable("%v", err)
	}
	return MakeSuccess("pem file updated")
}

func (worker *NamespaceWorker) delete(secret *api_v1.Secret) Status {
	if !worker.files.Have(secret.Namespace, secret.Name, string(secret.UID),
		secret.ResourceVersion) {
		return MakeNoop("pem file currently not stored")
	}

	exist, err := worker.files.Delete(secret.Namespace, secret.Name)
	if !exist {
		return MakeNoop("pem file not found")
	}
	if err != nil {
		if os.IsPermission(err) {
			return MakeFatal("%s", err)
		}
		return MakeRecoverable("%v", err)
	}
	return MakeSuccess("pem file deleted")
}

func (worker *NamespaceWorker) syncSuccess(syncType SyncType,
	secret *api_v1.Secret, status Status) {

	worker.log.Infof("%s Secret %s/%s: %s", syncType, secret.Namespace,
		secret.Name, status)
	worker.recorder.Eventf(secret, api_v1.EventTypeNormal, status.Reason(),
		"%s: %s", syncType, status)
}

func (worker *NamespaceWorker) syncFailure(syncType SyncType,
	secret *api_v1.Secret, status Status) {

	worker.log.Errorf("%s Secret %s/%s: %s", syncType, secret.Namespace,
		secret.Name, status)
	worker.recorder.Eventf(secret, api_v1.EventTypeWarning, status.Reason(),
		"%s: %s", syncType, status)
}

func (worker *NamespaceWorker) next() {
	obj, quit := worker.queue.Get()
	if quit {
		return
	}
	defer worker.queue.Done(obj)

	syncObj, ok := obj.(*SyncObj)
	if !ok {
		worker.log.Errorf("Unhandled type %T", obj)
		worker.queue.Forget(obj)
	}

	secret := syncObj.Secret
	if secret.Type != api_v1.SecretTypeTLS {
		worker.log.Panicf("Secret %s/%s: type is %s, should be %s",
			secret.Namespace, secret.Name, secret.Type,
			api_v1.SecretTypeTLS)
	}
	syncType := syncObj.Type

	var status Status
	switch syncObj.Type {
	case Update:
		status = worker.update(secret)
	case Delete:
		status = worker.delete(secret)
	default:
		status = MakeFatal("Unhandled sync: %s %s/%s", syncType,
			secret.Namespace, secret.Name)
	}

	switch status.Type {
	case Noop:
		worker.log.Infof("%s Secret %s/%s: %s", syncType,
			secret.Namespace, secret.Name, status)
		worker.queue.Forget(obj)
	case Success:
		worker.syncSuccess(syncType, secret, status)
		worker.queue.Forget(obj)
	case Fatal:
		worker.syncFailure(syncType, secret, status)
		worker.queue.Forget(obj)
	case Recoverable:
		worker.syncFailure(syncType, secret, status)
		worker.queue.AddRateLimited(obj)
	}
}

func (worker *NamespaceWorker) work() {
	defer worker.wg.Done()

	worker.log.Info("Starting worker for namespace:", worker.namespace)

	for !worker.queue.ShuttingDown() {
		worker.next()
	}

	worker.log.Info("Shutting down worker for namespace:", worker.namespace)
}

// NamespaceQueues reads from the main queue to which informers add
// new work items from all namespaces. The worker reads items from the
// queue and places them on separate queues for NamespaceWorkers
// responsible for each namespace.
type NamespaceQueues struct {
	Queue    workqueue.RateLimitingInterface
	log      *logrus.Logger
	workers  map[string]*NamespaceWorker
	files    *pem.Files
	recorder record.EventRecorder
	wg       *sync.WaitGroup
}

// NewNamespaceQueues creates a NamespaceQueues object.
//
//    log: logger initialized at startup
//    files: for managing PEM files
//    recorder: Event broadcaster initialized at startup
func NewNamespaceQueues(
	log *logrus.Logger,
	files *pem.Files,
	recorder record.EventRecorder,
) *NamespaceQueues {
	q := workqueue.NewRateLimitingQueue(
		workqueue.DefaultControllerRateLimiter())
	return &NamespaceQueues{
		Queue:    q,
		log:      log,
		workers:  make(map[string]*NamespaceWorker),
		files:    files,
		recorder: recorder,
		wg:       new(sync.WaitGroup),
	}
}

func (qs *NamespaceQueues) next() {
	obj, quit := qs.Queue.Get()
	if quit {
		return
	}
	defer qs.Queue.Done(obj)

	syncObj, ok := obj.(*SyncObj)
	if !ok {
		utilruntime.HandleError(fmt.Errorf("Unhandled type %T", obj))
		return
	}
	secret := syncObj.Secret
	worker, exists := qs.workers[secret.Namespace]
	if !exists {
		q := workqueue.NewNamedRateLimitingQueue(
			workqueue.DefaultControllerRateLimiter(),
			secret.Namespace)
		worker = &NamespaceWorker{
			namespace: secret.Namespace,
			log:       qs.log,
			queue:     q,
			files:     qs.files,
			recorder:  qs.recorder,
			wg:        qs.wg,
		}
		qs.workers[secret.Namespace] = worker
		qs.wg.Add(1)
		go worker.work()
	}
	worker.queue.Add(obj)
	qs.Queue.Forget(obj)
}

// Run comprises the main loop of the controller, reading from the
// main queue of work items and handing them off to workers for each
// namespace.
func (qs *NamespaceQueues) Run() {
	qs.log.Info("Starting dispatcher worker")
	for !qs.Queue.ShuttingDown() {
		qs.next()
	}
}

// Stop shuts down the main queue loop initiated by Run(), and in turn
// shuts down all of the NamespaceWorkers.
func (qs *NamespaceQueues) Stop() {
	qs.log.Info("Shutting down dispatcher worker")
	qs.Queue.ShutDown()
	for _, worker := range qs.workers {
		qs.log.Infof("Shutting down queue for namespace: %s",
			worker.namespace)
		worker.queue.ShutDown()
	}
	qs.log.Info("Waiting for workers to shut down")
	qs.wg.Wait()
}
