/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package pem

import (
	"encoding/json"
	"fmt"
	"os"
	"path/filepath"
	"reflect"
	"runtime"
	"sync"
	"testing"
	"time"
)

var basedir string

func TestMain(m *testing.M) {
	_, filename, _, ok := runtime.Caller(0)
	if !ok {
		fmt.Fprintln(os.Stderr, "Cannot get test directory")
		os.Exit(-1)
	}
	curDir := filepath.Dir(filename)
	basedir = filepath.Join(curDir, "testdata")

	os.Exit(m.Run())
}

const (
	whiskeyPem = `-----BEGIN CERTIFICATE-----
MIIFtTCCA52gAwIBAgIUL23Y5IPw50RNOsyXUg4Yk9elHw0wDQYJKoZIhvcNAQEL
BQAwajELMAkGA1UEBhMCREUxEDAOBgNVBAgMB0hhbWJ1cmcxEDAOBgNVBAcMB0hh
bWJ1cmcxHTAbBgNVBAoMFFRoZSBOZXh0IFdoaXNrZXkgQmFyMRgwFgYDVQQDDA9i
YXIuZXhhbXBsZS5jb20wHhcNMjAwNTEzMTI0NDMzWhcNNDAwNTA4MTI0NDMzWjBq
MQswCQYDVQQGEwJERTEQMA4GA1UECAwHSGFtYnVyZzEQMA4GA1UEBwwHSGFtYnVy
ZzEdMBsGA1UECgwUVGhlIE5leHQgV2hpc2tleSBCYXIxGDAWBgNVBAMMD2Jhci5l
eGFtcGxlLmNvbTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBAK4LQefd
TO7Q14fhxfWwZ3pM/VUCS1SEo2+efFx11kwXUY4h3R3uV8Ezx7advFnMWANsxA1s
NUNnVoPZ6jZ17DIp/oBbBHUbZ0m/UlH4frQluZX4RfbOtdHGjns1HtfZOkE7qW1J
qht2taLdCjxGob8vJc/EBfKtp4Tbl9U1rlk9E2rxk1KVpVoxgKrgnmbABYDb/yA+
KsDqHmPLnQiVeZ3CoflI9aFMAmcNnw18CfGfnUWyd5vVSgUCSyaw28bTwRJtWhWD
6fB9Rus2E5k5VaBqAL49bFharlV4zivfc2vcXdah1/W5+Vrp/vW/OgpU1p4s7454
kyItYZb2GGKxKVtH1tNatb3X+aY7aFfYTAiSeeCtpkMV1a14sA6B3jw3e7+UzwSc
E9/ljgV+h/JOtbxOexYRCZJel3OLxfKKq9YrdH2tUjCTWJhZH91wzRWLcidNlFL0
I1Q8xlu1Cif46piYTBkgyco4Mj3UHxSpQnVG4+3A7ZT5alemS4iukOZIUEs97CuI
RCC9UgKDHDzXE3GN0Qh7syWiyvqszdrCnB9iLNsv52eBC4DG1D65C9wY/1VdI+k1
6vJ+LqWh2aoKMcctLiq6o6z32uJff3FUkB8q8bWQ/OemDwtfWt9WW3qxRy9NvBlX
N7w1YejTlrAeIAEehZszgbBaDQXUaa60cngHAgMBAAGjUzBRMB0GA1UdDgQWBBTu
4JHYJOrVGPlooPrLZL2TzDeUhjAfBgNVHSMEGDAWgBTu4JHYJOrVGPlooPrLZL2T
zDeUhjAPBgNVHRMBAf8EBTADAQH/MA0GCSqGSIb3DQEBCwUAA4ICAQBvaUAyKqMv
1/77rX8bOkBSgscl2T50/Em4HOXo0WFTV6Lef3LIbNNu5xASf/TVX9Ckfrr42CKP
vS1B/HF27L+kB1cmd76BaaLd2mqJ9SYiww8y7N280IzoVUTc72kEzCVUtY/y9zCJ
olbt8zetZ0B4w5cba0TqRQYScDCAWmqnRUGF37IDSlvN3bNnoGOv8PVFsvswtn1l
pIKuyCyO1wCk7BPkdltLXysxe2m+cfIbosdCBKpCKj+iso1FqrPVXaoHiVHGvc9C
36vge9gNhR69sbrePbQrEB1mKp3HVf38qp0mlinOcNbwdRVxwaK33Q7kDO2w7JL9
oucFgd8w/HNqNU/HiemKPKjXrrJGQGQDltvtGEhnWLro8ez0bZZqANSnWLZdXpJX
84Lhb58bMuxBG9jnUc2wcmMbjiISpq8oGhajUAATnkc/B8B1vHZ73lNSdIUL61VA
o7lOZrYW6PSGh1QixHa7D1Nid5hcj6aaymNKyi7ESj5XTlbqJaBb+8zeVOR64HxJ
BFJG0FzRjk/TheVL8aO1Y7cj8woPcWGJj0ZJhBY6kuEN44nv7NbsXkiW4hJ1wHVQ
gWLNsQYCwyES3pIgliBkog54uFMGjpyeUJeATBcZpkvztjXS9GrVQhI6L3jEgN0C
4sa0SgvX/NR/L52KBSUWb4PX+VYWHw01nA==
-----END CERTIFICATE-----
`
	cafePem = `-----BEGIN CERTIFICATE-----
MIIDWTCCAkECFHb8EN0l0QwiR4eKKIW6h172z+JrMA0GCSqGSIb3DQEBCwUAMGgx
CzAJBgNVBAYTAkRFMRAwDgYDVQQIDAdIYW1idXJnMRAwDgYDVQQHDAdIYW1idXJn
MRowGAYDVQQKDBFHcmVlbiBNaWRnZXQgQ2FmZTEZMBcGA1UEAwwQY2FmZS5leGFt
cGxlLmNvbTAgFw0yMDA1MDQxNzA5NTlaGA8yMTIwMDQxMDE3MDk1OVowaDELMAkG
A1UEBhMCREUxEDAOBgNVBAgMB0hhbWJ1cmcxEDAOBgNVBAcMB0hhbWJ1cmcxGjAY
BgNVBAoMEUdyZWVuIE1pZGdldCBDYWZlMRkwFwYDVQQDDBBjYWZlLmV4YW1wbGUu
Y29tMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAts0HCq6fq9gv0uEa
3iOruZ3GnctdCoeGjrQQ4Fh2cQoMm/i3pkDUt6x2pLTQhxlN3oH3WEo1a24r/3S8
Xfy6Xf0Pti+dDiCqAwMd6veu56RItVMO1pmx1wDjGFTuplpnPRtz8EKsaKYfjZd1
BabdhkWhsA9g3nns8+lqeNbvebhk7hiv9lpgDWAnBie+hioan4WQdPZm1/bANH6o
+oWDu1o6Gdrk/iaj2pR73VTFsR2UEmSTpXa35W7/nsmgADIc4RovU+9ho1I4/fSy
jgVlZVBz29yLaDyNuoZljzNhvGqq1wW6Jq/v1uBOPxNH1k3ZQJl4jlG0tsoASnm7
mr9hewIDAQABMA0GCSqGSIb3DQEBCwUAA4IBAQBMNCVYMTdlaNaTjJ5Cznk9Gd+u
TSIFmOCetTOt3l0Xe0bSTxboT6Oz9nFDMP2A2HRK/GTp25ec+Ek1iiCIF47RcsGp
Cdug+x4wQVP3pxakJ/odFN1ReZGZCjNwBltxlRXwJhArK5PWmQppmMZPrW1UYW8y
x+m5UREzOzWga6EIlhpMEfgNa0BNCL/2gPaz2MpKXq5We93IDe2O0nlRrrVoDHU2
GFMhTpWSLkloaMzIMlcKR0IGyezG9waVgsliS00bYKp8eRJ5SqCUYvCMuApjoyzW
N2w59p6t5xE7Ktb0cmhZg83ISPTBlGqVJxF0clLob5nWyeutXNkP/KOi38PI
-----END CERTIFICATE-----
`

	ecdsaPem = `-----BEGIN CERTIFICATE-----
MIIBHjCBxaADAgECAgEBMAoGCCqGSM49BAMCMBcxFTATBgNVBAoTDERvY2tlciwg
SW5jLjAeFw0xMzA3MjUwMTEwMjRaFw0xNTA3MjUwMTEwMjRaMBcxFTATBgNVBAoT
DERvY2tlciwgSW5jLjBZMBMGByqGSM49AgEGCCqGSM49AwEHA0IABMolCWAO0iP7
tkX/KLjQ9CKeOoHYynBgfFcd1ZGoxcefmIbWjHx29eWI3xlhbjS6ssSxhrw1Kuh5
RrASfUCHD7SjAjAAMAoGCCqGSM49BAMCA0gAMEUCIQDRLQTSSeqjsxsb+q4exLSt
EM7f7/ymBzoUzbXU7wI9AgIgXCWaI++GkopGT8T2qV/3+NL0U+fYM0ZjSNSiwaK3
+kA=
-----END CERTIFICATE-----
`

	ed25519Pem = `-----BEGIN CERTIFICATE-----
MIIBLDCB36ADAgECAghWAUdKKo3DMDAFBgMrZXAwGTEXMBUGA1UEAwwOSUVURiBUZX
N0IERlbW8wHhcNMTYwODAxMTIxOTI0WhcNNDAxMjMxMjM1OTU5WjAZMRcwFQYDVQQD
DA5JRVRGIFRlc3QgRGVtbzAqMAUGAytlbgMhAIUg8AmJMKdUdIt93LQ+91oNvzoNJj
ga9OukqY6qm05qo0UwQzAPBgNVHRMBAf8EBTADAQEAMA4GA1UdDwEBAAQEAwIDCDAg
BgNVHQ4BAQAEFgQUmx9e7e0EM4Xk97xiPFl1uQvIuzswBQYDK2VwA0EAryMB/t3J5v
/BzKc9dNZIpDmAgs3babFOTQbs+BolzlDUwsPrdGxO3YNGhW7Ibz3OGhhlxXrCe1Cg
w1AH9efZBw==
-----END CERTIFICATE-----
`
)

var (
	expWhiskeyCrt = Crt{
		Version: 3,
		SN:      "2f:6d:d8:e4:83:f0:e7:44:4d:3a:cc:97:52:0e:18:93:d7:a5:1f:0d",
		Issuer:  "CN=bar.example.com,O=The Next Whiskey Bar,L=Hamburg,ST=Hamburg,C=DE",
		Valid: Validity{
			NotBefore: time.Date(2020, 5, 13, 12, 44, 33, 0, time.UTC),
			NotAfter:  time.Date(2040, 5, 8, 12, 44, 33, 0, time.UTC),
		},
		Subject: "CN=bar.example.com,O=The Next Whiskey Bar,L=Hamburg,ST=Hamburg,C=DE",
		PubKey: PubKeyInfo{
			Algorithm: "RSA",
			KeyInfo: RSAKeyInfo{
				Bits:     4096,
				Modulus:  "ae:0b:41:e7:dd:4c:ee:d0:d7:87:e1:c5:f5:b0:67:7a:4c:fd:55:02:4b:54:84:a3:6f:9e:7c:5c:75:d6:4c:17:51:8e:21:dd:1d:ee:57:c1:33:c7:b6:9d:bc:59:cc:58:03:6c:c4:0d:6c:35:43:67:56:83:d9:ea:36:75:ec:32:29:fe:80:5b:04:75:1b:67:49:bf:52:51:f8:7e:b4:25:b9:95:f8:45:f6:ce:b5:d1:c6:8e:7b:35:1e:d7:d9:3a:41:3b:a9:6d:49:aa:1b:76:b5:a2:dd:0a:3c:46:a1:bf:2f:25:cf:c4:05:f2:ad:a7:84:db:97:d5:35:ae:59:3d:13:6a:f1:93:52:95:a5:5a:31:80:aa:e0:9e:66:c0:05:80:db:ff:20:3e:2a:c0:ea:1e:63:cb:9d:08:95:79:9d:c2:a1:f9:48:f5:a1:4c:02:67:0d:9f:0d:7c:09:f1:9f:9d:45:b2:77:9b:d5:4a:05:02:4b:26:b0:db:c6:d3:c1:12:6d:5a:15:83:e9:f0:7d:46:eb:36:13:99:39:55:a0:6a:00:be:3d:6c:58:5a:ae:55:78:ce:2b:df:73:6b:dc:5d:d6:a1:d7:f5:b9:f9:5a:e9:fe:f5:bf:3a:0a:54:d6:9e:2c:ef:8e:78:93:22:2d:61:96:f6:18:62:b1:29:5b:47:d6:d3:5a:b5:bd:d7:f9:a6:3b:68:57:d8:4c:08:92:79:e0:ad:a6:43:15:d5:ad:78:b0:0e:81:de:3c:37:7b:bf:94:cf:04:9c:13:df:e5:8e:05:7e:87:f2:4e:b5:bc:4e:7b:16:11:09:92:5e:97:73:8b:c5:f2:8a:ab:d6:2b:74:7d:ad:52:30:93:58:98:59:1f:dd:70:cd:15:8b:72:27:4d:94:52:f4:23:54:3c:c6:5b:b5:0a:27:f8:ea:98:98:4c:19:20:c9:ca:38:32:3d:d4:1f:14:a9:42:75:46:e3:ed:c0:ed:94:f9:6a:57:a6:4b:88:ae:90:e6:48:50:4b:3d:ec:2b:88:44:20:bd:52:02:83:1c:3c:d7:13:71:8d:d1:08:7b:b3:25:a2:ca:fa:ac:cd:da:c2:9c:1f:62:2c:db:2f:e7:67:81:0b:80:c6:d4:3e:b9:0b:dc:18:ff:55:5d:23:e9:35:ea:f2:7e:2e:a5:a1:d9:aa:0a:31:c7:2d:2e:2a:ba:a3:ac:f7:da:e2:5f:7f:71:54:90:1f:2a:f1:b5:90:fc:e7:a6:0f:0b:5f:5a:df:56:5b:7a:b1:47:2f:4d:bc:19:57:37:bc:35:61:e8:d3:96:b0:1e:20:01:1e:85:9b:33:81:b0:5a:0d:05:d4:69:ae:b4:72:78:07",
				Exponent: 65537,
			},
		},
		SubjectKeyID:   "ee:e0:91:d8:24:ea:d5:18:f9:68:a0:fa:cb:64:bd:93:cc:37:94:86",
		AuthorityKeyID: "ee:e0:91:d8:24:ea:d5:18:f9:68:a0:fa:cb:64:bd:93:cc:37:94:86",
		Basic: BasicConstraints{
			Valid:   true,
			IsCA:    true,
			MaxPath: -1,
		},
		SigInfo: SigInfo{
			Algorithm: "sha256WithRSAEncryption",
			Signature: "6f:69:40:32:2a:a3:2f:d7:fe:fb:ad:7f:1b:3a:40:52:82:c7:25:d9:3e:74:fc:49:b8:1c:e5:e8:d1:61:53:57:a2:de:7f:72:c8:6c:d3:6e:e7:10:12:7f:f4:d5:5f:d0:a4:7e:ba:f8:d8:22:8f:bd:2d:41:fc:71:76:ec:bf:a4:07:57:26:77:be:81:69:a2:dd:da:6a:89:f5:26:22:c3:0f:32:ec:dd:bc:d0:8c:e8:55:44:dc:ef:69:04:cc:25:54:b5:8f:f2:f7:30:89:a2:56:ed:f3:37:ad:67:40:78:c3:97:1b:6b:44:ea:45:06:12:70:30:80:5a:6a:a7:45:41:85:df:b2:03:4a:5b:cd:dd:b3:67:a0:63:af:f0:f5:45:b2:fb:30:b6:7d:65:a4:82:ae:c8:2c:8e:d7:00:a4:ec:13:e4:76:5b:4b:5f:2b:31:7b:69:be:71:f2:1b:a2:c7:42:04:aa:42:2a:3f:a2:b2:8d:45:aa:b3:d5:5d:aa:07:89:51:c6:bd:cf:42:df:ab:e0:7b:d8:0d:85:1e:bd:b1:ba:de:3d:b4:2b:10:1d:66:2a:9d:c7:55:fd:fc:aa:9d:26:96:29:ce:70:d6:f0:75:15:71:c1:a2:b7:dd:0e:e4:0c:ed:b0:ec:92:fd:a2:e7:05:81:df:30:fc:73:6a:35:4f:c7:89:e9:8a:3c:a8:d7:ae:b2:46:40:64:03:96:db:ed:18:48:67:58:ba:e8:f1:ec:f4:6d:96:6a:00:d4:a7:58:b6:5d:5e:92:57:f3:82:e1:6f:9f:1b:32:ec:41:1b:d8:e7:51:cd:b0:72:63:1b:8e:22:12:a6:af:28:1a:16:a3:50:00:13:9e:47:3f:07:c0:75:bc:76:7b:de:53:52:74:85:0b:eb:55:40:a3:b9:4e:66:b6:16:e8:f4:86:87:54:22:c4:76:bb:0f:53:62:77:98:5c:8f:a6:9a:ca:63:4a:ca:2e:c4:4a:3e:57:4e:56:ea:25:a0:5b:fb:cc:de:54:e4:7a:e0:7c:49:04:52:46:d0:5c:d1:8e:4f:d3:85:e5:4b:f1:a3:b5:63:b7:23:f3:0a:0f:71:61:89:8f:46:49:84:16:3a:92:e1:0d:e3:89:ef:ec:d6:ec:5e:48:96:e2:12:75:c0:75:50:81:62:cd:b1:06:02:c3:21:12:de:92:20:96:20:64:a2:0e:78:b8:53:06:8e:9c:9e:50:97:80:4c:17:19:a6:4b:f3:b6:35:d2:f4:6a:d5:42:12:3a:2f:78:c4:80:dd:02:e2:c6:b4:4a:0b:d7:fc:d4:7f:2f:9d:8a:05:25:16:6f:83:d7:f9:56:16:1f:0d:35:9c",
		},
		PEM: whiskeyPem,
	}

	expCafeCrt = Crt{
		Version: 1,
		SN:      "76:fc:10:dd:25:d1:0c:22:47:87:8a:28:85:ba:87:5e:f6:cf:e2:6b",
		Issuer:  "CN=cafe.example.com,O=Green Midget Cafe,L=Hamburg,ST=Hamburg,C=DE",
		Valid: Validity{
			NotBefore: time.Date(2020, 5, 4, 17, 9, 59, 0, time.UTC),
			NotAfter:  time.Date(2120, 4, 10, 17, 9, 59, 0, time.UTC),
		},
		Subject: "CN=cafe.example.com,O=Green Midget Cafe,L=Hamburg,ST=Hamburg,C=DE",
		PubKey: PubKeyInfo{
			Algorithm: "RSA",
			KeyInfo: RSAKeyInfo{
				Bits:     2048,
				Modulus:  "b6:cd:07:0a:ae:9f:ab:d8:2f:d2:e1:1a:de:23:ab:b9:9d:c6:9d:cb:5d:0a:87:86:8e:b4:10:e0:58:76:71:0a:0c:9b:f8:b7:a6:40:d4:b7:ac:76:a4:b4:d0:87:19:4d:de:81:f7:58:4a:35:6b:6e:2b:ff:74:bc:5d:fc:ba:5d:fd:0f:b6:2f:9d:0e:20:aa:03:03:1d:ea:f7:ae:e7:a4:48:b5:53:0e:d6:99:b1:d7:00:e3:18:54:ee:a6:5a:67:3d:1b:73:f0:42:ac:68:a6:1f:8d:97:75:05:a6:dd:86:45:a1:b0:0f:60:de:79:ec:f3:e9:6a:78:d6:ef:79:b8:64:ee:18:af:f6:5a:60:0d:60:27:06:27:be:86:2a:1a:9f:85:90:74:f6:66:d7:f6:c0:34:7e:a8:fa:85:83:bb:5a:3a:19:da:e4:fe:26:a3:da:94:7b:dd:54:c5:b1:1d:94:12:64:93:a5:76:b7:e5:6e:ff:9e:c9:a0:00:32:1c:e1:1a:2f:53:ef:61:a3:52:38:fd:f4:b2:8e:05:65:65:50:73:db:dc:8b:68:3c:8d:ba:86:65:8f:33:61:bc:6a:aa:d7:05:ba:26:af:ef:d6:e0:4e:3f:13:47:d6:4d:d9:40:99:78:8e:51:b4:b6:ca:00:4a:79:bb:9a:bf:61:7b",
				Exponent: 65537,
			},
		},
		Basic: BasicConstraints{
			Valid:   false,
			IsCA:    false,
			MaxPath: -1,
		},
		SigInfo: SigInfo{
			Algorithm: "sha256WithRSAEncryption",
			Signature: "4c:34:25:58:31:37:65:68:d6:93:8c:9e:42:ce:79:3d:19:df:ae:4d:22:05:98:e0:9e:b5:33:ad:de:5d:17:7b:46:d2:4f:16:e8:4f:a3:b3:f6:71:43:30:fd:80:d8:74:4a:fc:64:e9:db:97:9c:f8:49:35:8a:20:88:17:8e:d1:72:c1:a9:09:db:a0:fb:1e:30:41:53:f7:a7:16:a4:27:fa:1d:14:dd:51:79:91:99:0a:33:70:06:5b:71:95:15:f0:26:10:2b:2b:93:d6:99:0a:69:98:c6:4f:ad:6d:54:61:6f:32:c7:e9:b9:51:11:33:3b:35:a0:6b:a1:08:96:1a:4c:11:f8:0d:6b:40:4d:08:bf:f6:80:f6:b3:d8:ca:4a:5e:ae:56:7b:dd:c8:0d:ed:8e:d2:79:51:ae:b5:68:0c:75:36:18:53:21:4e:95:92:2e:49:68:68:cc:c8:32:57:0a:47:42:06:c9:ec:c6:f7:06:95:82:c9:62:4b:4d:1b:60:aa:7c:79:12:79:4a:a0:94:62:f0:8c:b8:0a:63:a3:2c:d6:37:6c:39:f6:9e:ad:e7:11:3b:2a:d6:f4:72:68:59:83:cd:c8:48:f4:c1:94:6a:95:27:11:74:72:52:e8:6f:99:d6:c9:eb:ad:5c:d9:0f:fc:a3:a2:df:c3:c8",
		},
		PEM: cafePem,
	}

	expEcdsaCrt = Crt{
		Version: 3,
		SN:      "01",
		Issuer:  "O=Docker\\, Inc.",
		Valid: Validity{
			NotBefore: time.Date(2013, 7, 25, 1, 10, 24, 0, time.UTC),
			NotAfter:  time.Date(2015, 7, 25, 1, 10, 24, 0, time.UTC),
		},
		Subject: "O=Docker\\, Inc.",
		PubKey: PubKeyInfo{
			Algorithm: "ECDSA",
			KeyInfo: ECDSAKeyInfo{
				Name: "P-256",
				X:    "ca:25:09:60:0e:d2:23:fb:b6:45:ff:28:b8:d0:f4:22:9e:3a:81:d8:ca:70:60:7c:57:1d:d5:91:a8:c5:c7:9f",
				Y:    "98:86:d6:8c:7c:76:f5:e5:88:df:19:61:6e:34:ba:b2:c4:b1:86:bc:35:2a:e8:79:46:b0:12:7d:40:87:0f:b4",
			},
		},
		Basic: BasicConstraints{
			Valid:   false,
			IsCA:    false,
			MaxPath: -1,
		},
		SigInfo: SigInfo{
			Algorithm: "ecdsaWithSHA256",
			Signature: "30:45:02:21:00:d1:2d:04:d2:49:ea:a3:b3:1b:1b:fa:ae:1e:c4:b4:ad:10:ce:df:ef:fc:a6:07:3a:14:cd:b5:d4:ef:02:3d:02:02:20:5c:25:9a:23:ef:86:92:8a:46:4f:c4:f6:a9:5f:f7:f8:d2:f4:53:e7:d8:33:46:63:48:d4:a2:c1:a2:b7:fa:40",
		},
		PEM: ecdsaPem,
	}

	expEd25519Crt = Crt{
		Version: 3,
		SN:      "56:01:47:4a:2a:8d:c3:30",
		Issuer:  "CN=IETF Test Demo",
		Valid: Validity{
			NotBefore: time.Date(2016, 8, 1, 12, 19, 24, 0, time.UTC),
			NotAfter:  time.Date(2040, 12, 31, 23, 59, 59, 0, time.UTC),
		},
		Subject: "CN=IETF Test Demo",
		PubKey: PubKeyInfo{
			Algorithm: "unknown",
			KeyInfo:   UnknownKeyInfo{},
		},
		SubjectKeyID: "9b:1f:5e:ed:ed:04:33:85:e4:f7:bc:62:3c:59:75:b9:0b:c8:bb:3b",
		Basic: BasicConstraints{
			Valid:   true,
			IsCA:    false,
			MaxPath: -1,
		},
		KeyUsage: []string{"keyAgreement"},
		SigInfo: SigInfo{
			Algorithm: "unknown",
			Signature: "af:23:01:fe:dd:c9:e6:ff:c1:cc:a7:3d:74:d6:48:a4:39:80:82:cd:db:69:b1:4e:4d:06:ec:f8:1a:25:ce:50:d4:c2:c3:eb:74:6c:4e:dd:83:46:85:6e:c8:6f:3d:ce:1a:18:65:c5:7a:c2:7b:50:a0:c3:50:07:f5:e7:d9:07",
		},
		PEM: ed25519Pem,
	}
)

func TestMkCrt(t *testing.T) {
	crt, err := mkCrt([]byte(whiskeyPem))
	if err != nil {
		t.Fatalf("mkCrt(whiskeyPem): %v", err)
	}
	if !reflect.DeepEqual(crt, expWhiskeyCrt) {
		t.Errorf("mkCrt(whiskyPem) got %#v want %#v", crt,
			expWhiskeyCrt)
	}
	jsonBytes, err := json.Marshal(crt)
	if err != nil {
		t.Fatalf("json.Marshal(crt): %v", err)
	}
	if err = json.Unmarshal(jsonBytes, &crt); err != nil {
		t.Fatalf("json.Unmarshal(whiskeyJson): %v", err)
	}
	if !reflect.DeepEqual(crt, expWhiskeyCrt) {
		t.Errorf("json.Unmarshal(whiskeyJson) got %#v want %#v", crt,
			expWhiskeyCrt)
	}

	crt, err = mkCrt([]byte(cafePem))
	if err != nil {
		t.Fatalf("mkCrt(cafePem): %v", err)
	}
	if !reflect.DeepEqual(crt, expCafeCrt) {
		t.Errorf("mkCrt(cafePem) got %#v want %#v", crt, expCafeCrt)
	}
	jsonBytes, err = json.Marshal(crt)
	if err != nil {
		t.Fatalf("json.Marshal(crt): %v", err)
	}
	if err = json.Unmarshal(jsonBytes, &crt); err != nil {
		t.Fatalf("json.Unmarshal(cafeJson): %v", err)
	}
	if !reflect.DeepEqual(crt, expCafeCrt) {
		t.Errorf("json.Unmarshal(cafeJson) got %#v want %#v", crt,
			expCafeCrt)
	}

	crt, err = mkCrt([]byte(ecdsaPem))
	if err != nil {
		t.Fatalf("mkCrt(ecdsaPem): %v", err)
	}
	if !reflect.DeepEqual(crt, expEcdsaCrt) {
		t.Errorf("mkCrt(ecdsaPem) got %#v want %#v", crt, expEcdsaCrt)
	}
	jsonBytes, err = json.Marshal(crt)
	if err != nil {
		t.Fatalf("json.Marshal(crt): %v", err)
	}
	if err = json.Unmarshal(jsonBytes, &crt); err != nil {
		t.Fatalf("json.Unmarshal(ecdsaJson): %v", err)
	}
	if !reflect.DeepEqual(crt, expEcdsaCrt) {
		t.Errorf("json.Unmarshal(ecdsaJson) got %#v want %#v", crt,
			expEcdsaCrt)
	}

	crt, err = mkCrt([]byte(ed25519Pem))
	if err != nil {
		t.Fatalf("mkCrt(ed25519Pem): %v", err)
	}
	if !reflect.DeepEqual(crt, expEd25519Crt) {
		t.Errorf("mkCrt(ed25519Pem) got %#v want %#v", crt,
			expEd25519Crt)
	}
	jsonBytes, err = json.Marshal(crt)
	if err != nil {
		t.Fatalf("json.Marshal(crt): %v", err)
	}
	if err = json.Unmarshal(jsonBytes, &crt); err != nil {
		t.Fatalf("json.Unmarshal(ed25519Json): %v", err)
	}
	if !reflect.DeepEqual(crt, expEd25519Crt) {
		t.Errorf("json.Unmarshal(ed25519Json) got %#v want %#v", crt,
			expEd25519Crt)
	}
}

var fileInfo = FileInfo{
	K8sSecret: Secret{
		Namespace: "ns",
		Name:      "name",
		UID:       "1dc443a3-e23d-49e6-b016-fc8b1eaeb724",
		Version:   "4711",
	},
	Meta: FileMeta{
		ModTime: time.Date(2020, 7, 27, 9, 40, 26, 0, time.UTC),
		Size:    1147,
	},
	Cert: expWhiskeyCrt,
}

func TestFileInfoJSON(t *testing.T) {
	jsonBytes, err := json.Marshal(fileInfo)
	if err != nil {
		t.Fatalf("json.Marshal(fileInfo): %v", err)
	}
	var unmarshalledFileInfo FileInfo
	if err = json.Unmarshal(jsonBytes, &unmarshalledFileInfo); err != nil {
		t.Fatalf("json.Unmarshal(fileInfoJSON): %v", err)
	}
	if !reflect.DeepEqual(unmarshalledFileInfo, fileInfo) {
		t.Errorf("json.Unmarshal(fileInfoJSON) got %#v want %#v",
			unmarshalledFileInfo, fileInfo)
	}
}

var pemFiles = &Files{
	Files: map[string]*File{
		"ns1/cafe": &File{
			Namespace:       "ns1",
			Name:            "cafe",
			UID:             "975d4e4f-9ea9-49d9-a81d-1b4bca92a743",
			ResourceVersion: "4711",
			Size:            2899,
			ModTime:         time.Date(2020, 7, 27, 9, 50, 0, 0, time.UTC),
		},
		"ns2/bar": &File{
			Namespace:       "ns2",
			Name:            "bar",
			UID:             "d18974c5-94d7-4e04-b2af-6e9274ad46d8",
			ResourceVersion: "0815",
			Size:            5313,
			ModTime:         time.Date(2020, 7, 27, 9, 51, 0, 0, time.UTC),
		},
	},
	Gid: -1,
	mtx: new(sync.RWMutex),
}

var invalidFiles = &Files{
	Files: map[string]*File{
		"invalid/enoent": &File{
			Namespace: "invalid",
			Name:      "enoent",
		},
		"invalid/crt": &File{
			Namespace: "invalid",
			Name:      "crt",
		},
	},
	Gid: -1,
	mtx: new(sync.RWMutex),
}

var (
	expCafeInfo = FileInfo{
		K8sSecret: Secret{
			Namespace: "ns1",
			Name:      "cafe",
			UID:       "975d4e4f-9ea9-49d9-a81d-1b4bca92a743",
			Version:   "4711",
		},
		Meta: FileMeta{
			Size:    2899,
			ModTime: time.Date(2020, 7, 27, 9, 50, 0, 0, time.UTC),
		},
		Cert: expCafeCrt,
	}

	expBarInfo = FileInfo{
		K8sSecret: Secret{
			Namespace: "ns2",
			Name:      "bar",
			UID:       "d18974c5-94d7-4e04-b2af-6e9274ad46d8",
			Version:   "0815",
		},
		Meta: FileMeta{
			Size:    5313,
			ModTime: time.Date(2020, 7, 27, 9, 51, 0, 0, time.UTC),
		},
		Cert: expWhiskeyCrt,
	}
)

func TestGetFileInfo(t *testing.T) {
	pemFiles.Base = basedir

	info, found, err := pemFiles.GetFileInfo("ns1", "cafe", "", "")
	if !found {
		t.Fatal("GetFileInfo(ns1/cafe) found got false want true")
	}
	if err != nil {
		t.Fatalf("GetFileInfo(ns1/cafe): %v", err)
	}
	if !reflect.DeepEqual(info, expCafeInfo) {
		t.Errorf("GetFileInfo(ns1/cafe) got %#v want %#v",
			info, expCafeInfo)
	}
	jsonBytes, err := json.Marshal(info)
	if err != nil {
		t.Fatalf("json.Marshal(info): %v", err)
	}
	var unmarshalledFileInfo FileInfo
	if err = json.Unmarshal(jsonBytes, &unmarshalledFileInfo); err != nil {
		t.Fatalf("json.Unmarshal(fileInfoJSON): %v", err)
	}
	if !reflect.DeepEqual(unmarshalledFileInfo, info) {
		t.Errorf("json.Unmarshal(fileInfoJSON) got %#v want %#v",
			unmarshalledFileInfo, info)
	}

	info, found, err = pemFiles.GetFileInfo("ns2", "bar", "", "")
	if !found {
		t.Fatal("GetFileInfo(ns2/bar) found got false want true")
	}
	if err != nil {
		t.Fatalf("GetFileInfo(ns2/bar): %v", err)
	}
	if !reflect.DeepEqual(info, expBarInfo) {
		t.Errorf("GetFileInfo(ns2/bar) got %#v want %#v",
			info, expBarInfo)
	}
	jsonBytes, err = json.Marshal(info)
	if err != nil {
		t.Fatalf("json.Marshal(info): %v", err)
	}
	if err = json.Unmarshal(jsonBytes, &unmarshalledFileInfo); err != nil {
		t.Fatalf("json.Unmarshal(fileInfoJSON): %v", err)
	}
	if !reflect.DeepEqual(unmarshalledFileInfo, info) {
		t.Errorf("json.Unmarshal(fileInfoJSON) got %#v want %#v",
			unmarshalledFileInfo, info)
	}

	if info, found, err = pemFiles.GetFileInfo("ns1", "bar", "", ""); found {
		t.Error("GetFileInfo(ns1/bar) found got true want false")
	}
	if err == nil {
		t.Error("GetFileInfo(ns1/bar) error got nil want not found")
	}
	if info, found, err = pemFiles.GetFileInfo("ns2", "cafe", "", ""); found {
		t.Error("GetFileInfo(ns2/cafe) found got true want false")
	}
	if err == nil {
		t.Error("GetFileInfo(ns2/cafe) error got nil want not found")
	}
	if info, found, err = pemFiles.GetFileInfo("ns1", "cafe",
		"975d4e4f-9ea9-49d9-a81d-1b4bca92a743", ""); !found {
		t.Error("GetFileInfo(ns1/cafe,uid) found got false want true")
	}
	if err != nil {
		t.Errorf("GetFileInfo(ns1/cafe,uid) error got %v want nil", err)
	}
	if info, found, err = pemFiles.GetFileInfo(
		"ns1", "cafe", "", "4711"); !found {
		t.Error("GetFileInfo(ns1/cafe,version) found got false " +
			"want true")
	}
	if err != nil {
		t.Errorf("GetFileInfo(ns1/cafe,version) error got %v want nil",
			err)
	}
	if info, found, err = pemFiles.GetFileInfo("ns1", "cafe",
		"975d4e4f-9ea9-49d9-a81d-1b4bca92a743", "4711"); !found {
		t.Error("GetFileInfo(ns1/cafe,uid,version) found got false " +
			"want true")
	}
	if err != nil {
		t.Errorf("GetFileInfo(ns1/cafe,uid,version) error got %v "+
			"want nil", err)
	}
	if info, found, err = pemFiles.GetFileInfo("ns1", "cafe",
		"d18974c5-94d7-4e04-b2af-6e9274ad46d8", ""); found {
		t.Error("GetFileInfo(ns1/cafe,uid) found got true want false")
	}
	if err == nil {
		t.Error("GetFileInfo(ns1/cafe,uid) error got nil " +
			"want not found")
	}
	if info, found, err = pemFiles.GetFileInfo(
		"ns1", "cafe", "", "0815"); found {
		t.Error("GetFileInfo(ns1/cafe,version) found got true " +
			"want false")
	}
	if err == nil {
		t.Error("GetFileInfo(ns1/cafe,version) error got nil " +
			"want ono-nil")
	}
	if info, found, err = pemFiles.GetFileInfo("ns1", "cafe",
		"d18974c5-94d7-4e04-b2af-6e9274ad46d8", "0815"); found {
		t.Error("GetFileInfo(ns1/cafe,uid,version) error got true " +
			"want false")
	}
	if err == nil {
		t.Error("GetFileInfo(ns1/cafe,uid,version) error got nil " +
			"want not found")
	}

	invalidFiles.Base = basedir
	if _, found, err = invalidFiles.GetFileInfo(
		"invalid", "enoent", "", ""); found {
		t.Error("GetFileInfo(invalid/enoent) found got true " +
			"want false")
	}
	if err == nil {
		t.Errorf("GetFileInfo(invalid/enoent) error got nil " +
			"want not found")
	}
	if _, found, err = invalidFiles.GetFileInfo(
		"invalid", "crt", "", ""); !found {
		t.Error("GetFileInfo(invalid/crt) found got false " +
			"want true")
	}
	if err == nil {
		t.Error("GetFileInfo(invalid/crt) error got nil " +
			"want crt parse error")
	}
}

func TestGetAllFileInfo(t *testing.T) {
	pemFiles.Base = basedir

	allInfo, err := pemFiles.GetAllFileInfo()
	if err != nil {
		t.Fatalf("GetAllFileInfo(): %v", err)
	}
	if len(allInfo) != 2 {
		t.Fatalf("GetAllFileInfo() len(slice) got %d want 2",
			len(allInfo))
	}
	cafeFound, barFound := false, false
	for _, info := range allInfo {
		secret := info.K8sSecret
		if secret.Namespace == "ns1" && secret.Name == "cafe" {
			cafeFound = true
			if !reflect.DeepEqual(info, expCafeInfo) {
				t.Errorf("GetAllFileInfo() cafe got %#v "+
					"want %#v", info, expCafeInfo)
			}
		} else if secret.Namespace == "ns2" && secret.Name == "bar" {
			barFound = true
			if !reflect.DeepEqual(info, expBarInfo) {
				t.Errorf("GetAllFileInfo() bar got %#v "+
					"want %#v", info, expBarInfo)
			}
		} else {
			t.Errorf("GetAllFileInfo() unexpected: %#v", info)
		}
	}
	if !cafeFound {
		t.Errorf("GetAllFileInfo() did not return cafe: %#v", allInfo)
	}
	if !barFound {
		t.Errorf("GetAllFileInfo() did not return bar: %#v", allInfo)
	}
	jsonBytes, err := json.Marshal(allInfo)
	if err != nil {
		t.Fatalf("json.Marshal(allInfo): %v", err)
	}
	var unmarshalledAllInfo []FileInfo
	if err = json.Unmarshal(jsonBytes, &unmarshalledAllInfo); err != nil {
		t.Fatalf("json.Unmarshal(allInfoJSON): %v", err)
	}
	if !reflect.DeepEqual(unmarshalledAllInfo, allInfo) {
		t.Errorf("json.Unmarshal(allInfoJSON) got %#v want %#v",
			unmarshalledAllInfo, allInfo)
	}

	invalidFiles.Base = basedir
	if _, err = invalidFiles.GetAllFileInfo(); err == nil {
		t.Errorf("invalid.GetAllFileInfo() error got nil want non-nil")
	}
}
